@@ -1,14 +1,15 @@
-from __future__ import absolute_import
-import collections
+# -*- coding: utf-8 -*-
+import asyncio
+import collections.abc
+import functools
 import inspect
 import os
 import socket
-import tornado.concurrent
-from tornado import gen, ioloop
+import traceback
 
-import kiwipy
+from kiwipy import exceptions
 
-__all__ = []
+__all__ = ()
 
 # The key used in messages to give information about the host that send a message
 HOST_KEY = 'host'
@@ -26,7 +27,7 @@ def get_host_info():
 
 def add_host_info(msg):
     if HOST_KEY in msg:
-        raise ValueError("Host information key already exists in message")
+        raise ValueError('Host information key already exists in message')
 
     msg[HOST_KEY] = get_host_info()
 
@@ -35,18 +36,16 @@ def result_response(result):
     return {RESULT_KEY: result}
 
 
-def exception_response(exception, trace=None):
+def exception_response(exception: Exception, trace=None) -> dict:
     """
     Create an exception response dictionary
     :param exception: The exception to encode
-    :type exception: :class:`Exception`
     :param trace: Optional traceback
-    :return: A response dictionary
-    :rtype: dict
+    :return: An exception response dictionary
     """
     msg = str(exception)
     if trace is not None:
-        msg += "\n{}".format(trace)
+        msg += '\n{}'.format(''.join(traceback.format_tb(trace)[0]))
     return {EXCEPTION_KEY: msg}
 
 
@@ -59,7 +58,7 @@ def pending_response(msg=None):
 
 
 def response_result(response):
-    future = tornado.concurrent.Future()
+    future = asyncio.Future()
     response_to_future(response, future)
     return future.result()
 
@@ -71,20 +70,20 @@ def response_to_future(response, future=None):
     :param future:
     :return:
     """
-    if not isinstance(response, collections.Mapping):
-        raise TypeError("Response must be a mapping")
+    if not isinstance(response, collections.abc.Mapping):
+        raise TypeError('Response must be a mapping')
 
     if future is None:
-        future = tornado.concurrent.Future()
+        future = asyncio.Future()
 
     if CANCELLED_KEY in response:
         future.cancel()
     elif EXCEPTION_KEY in response:
-        future.set_exception(kiwipy.RemoteException(response[EXCEPTION_KEY]))
+        future.set_exception(exceptions.RemoteException(response[EXCEPTION_KEY]))
     elif RESULT_KEY in response:
         future.set_result(response[RESULT_KEY])
     elif PENDING_KEY in response:
-        future.set_result(tornado.concurrent.Future())
+        future.set_result(asyncio.Future())
     else:
         raise ValueError("Unknown response type '{}'".format(response))
 
@@ -108,42 +107,16 @@ def future_to_response(future):
 
 
 def ensure_coroutine(coro_or_fn):
-    if gen.is_coroutine_function(coro_or_fn):
+    if asyncio.iscoroutinefunction(coro_or_fn):
         return coro_or_fn
     if callable(coro_or_fn):
         if inspect.isclass(coro_or_fn):
             coro_or_fn = coro_or_fn.__call__
-        return gen.coroutine(coro_or_fn)
+
+        @functools.wraps(coro_or_fn)
+        async def wrap(*args, **kwargs):
+            return coro_or_fn(*args, **kwargs)
+
+        return wrap
 
     raise TypeError('coro_or_fn must be a callable')
-
-
-@gen.convert_yielded.register(kiwipy.Future)
-def _register_concurrent_future(concurrent_future):
-    """Register concurrent.future.Future to be compatible with tornado yield"""
-    tornado_future = tornado.concurrent.Future()
-    tornado.concurrent.chain_future(concurrent_future, tornado_future)
-    return tornado_future
-
-
-def create_task(coro, loop=None):
-    """
-    Schedule a call to a coroutine in the event loop and wrap the outcome
-    in a future.
-
-    :param coro: the coroutine to schedule
-    :param loop: the event loop to schedule it in
-    :return: the future representing the outcome of the coroutine
-    :rtype: :class:`tornado.concurrent.Future`
-    """
-    loop = loop or ioloop.IOLoop.current()
-
-    future = tornado.concurrent.Future()
-
-    @gen.coroutine
-    def run_task():
-        with kiwipy.capture_exceptions(future):
-            future.set_result((yield coro()))
-
-    loop.add_callback(run_task)
-    return future
