@@ -1,7 +1,8 @@
-from __future__ import absolute_import
-import kiwipy
+# -*- coding: utf-8 -*-
 import unittest
 
+import kiwipy
+
 from .utils import CommunicatorTester
 
 
