@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 """
 A generic test for any communicator that conforms to the `kiwipy.Communicator` interface.
 
@@ -5,26 +6,26 @@ I've had to disable invalid-name for some portions of this file as I use the con
 that all caps 'variable' names mean that it's a constant.
 """
 
-from __future__ import absolute_import
 import abc
-import six
+import re
+
+import pytest
 
 import kiwipy
 
 
-@six.add_metaclass(abc.ABCMeta)
-class CommunicatorTester(object):
+class CommunicatorTester(metaclass=abc.ABCMeta):
     # Disable invalid name because I use caps for constants which the linter doesn't like
     # also disable no-member for superclass calls because we use this as a mixin that gets used
     # with unittest.TestCase
-    # pylint: disable=invalid-name, no-member
+    # pylint: disable=invalid-name, no-member, too-many-public-methods
 
     WAIT_TIMEOUT = 5.
 
     communicator = None
 
     def setUp(self):
-        super(CommunicatorTester, self).setUp()
+        super().setUp()
         self.communicator = self.create_communicator()
 
     def tearDown(self):
@@ -37,16 +38,38 @@ class CommunicatorTester(object):
         :return: A constructed communicator
         :rtype: :class:`kiwi.Communicator`
         """
-        pass
 
     def destroy_communicator(self, communicator):
         pass
 
+    def test_close(self):
+        """Make sure a closed communicator raises when trying to perform any communication
+        operation"""
+        self.communicator.close()
+        with self.assertRaises(kiwipy.CommunicatorClosed):
+            self.communicator.add_rpc_subscriber(None, 'rpc')
+        with self.assertRaises(kiwipy.CommunicatorClosed):
+            self.communicator.rpc_send('rpc', None).result(timeout=self.WAIT_TIMEOUT)
+        with self.assertRaises(kiwipy.CommunicatorClosed):
+            self.communicator.add_task_subscriber(None)
+        with self.assertRaises(kiwipy.CommunicatorClosed):
+            self.communicator.task_send(None).result(timeout=self.WAIT_TIMEOUT)
+        with self.assertRaises(kiwipy.CommunicatorClosed):
+            self.communicator.add_broadcast_subscriber(None)
+        with self.assertRaises(kiwipy.CommunicatorClosed):
+            self.communicator.broadcast_send(None)
+
+    def test_context(self):
+        assert not self.communicator.is_closed()
+        with self.communicator:
+            pass
+        assert self.communicator.is_closed()
+
     # region RPC
 
     def test_rpc_send_receive(self):
         MESSAGE = "sup yo'"  # pylint: disable=invalid-name
-        RESPONSE = "nuthin bra"  # pylint: disable=invalid-name
+        RESPONSE = 'nuthin bra'  # pylint: disable=invalid-name
 
         messages = []
 
@@ -104,7 +127,7 @@ class CommunicatorTester(object):
 
     def test_rpc_exception(self):
         MESSAGE = 'The meaning?'  # pylint: disable=invalid-name
-        EXCEPTION = RuntimeError("I cannea do it Captain!")  # pylint: disable=invalid-name
+        EXCEPTION = RuntimeError('I cannae do it Captain!')  # pylint: disable=invalid-name
 
         messages = []
 
@@ -169,7 +192,7 @@ class CommunicatorTester(object):
 
     def test_task_exception(self):
         TASK = 'The meaning?'  # pylint: disable=invalid-name
-        EXCEPTION = RuntimeError("I cannea do it Captain!")  # pylint: disable=invalid-name
+        EXCEPTION = RuntimeError('I cannea do it Captain!')  # pylint: disable=invalid-name
 
         tasks = []
 
@@ -208,6 +231,22 @@ class CommunicatorTester(object):
         self.assertEqual(TASK, tasks[0])
         self.assertEqual(RESULT, result)
 
+    def test_add_remove_task_subscriber(self):
+
+        def worker(*_, **__):
+            pass
+
+        worker_id = self.communicator.add_task_subscriber(worker)
+        self.communicator.remove_task_subscriber(worker_id)
+
+        fut = self.communicator.task_send('task')
+        with pytest.raises(kiwipy.TimeoutError):
+            fut.result(0.01)
+
+    def test_remove_unknown_task_subscriber(self):
+        with pytest.raises(ValueError):
+            self.communicator.remove_task_subscriber('unknown')
+
     # endregion
 
     # region Broadcast
@@ -248,7 +287,7 @@ class CommunicatorTester(object):
             if len(subjects) == len(EXPECTED_SUBJECTS):
                 done.set_result(True)
 
-        self.communicator.add_broadcast_subscriber(kiwipy.BroadcastFilter(on_broadcast_1, subject="purchase.*"))
+        self.communicator.add_broadcast_subscriber(kiwipy.BroadcastFilter(on_broadcast_1, subject='purchase.*'))
 
         for subj in ['purchase.car', 'purchase.piano', 'sell.guitar', 'sell.house']:
             self.communicator.broadcast_send(None, subject=subj)
@@ -269,10 +308,10 @@ class CommunicatorTester(object):
             if len(senders) == len(EXPECTED_SENDERS):
                 done.set_result(True)
 
-        self.communicator.add_broadcast_subscriber(kiwipy.BroadcastFilter(on_broadcast_1, sender="*.jones"))
+        self.communicator.add_broadcast_subscriber(kiwipy.BroadcastFilter(on_broadcast_1, sender='*.jones'))
 
-        for sendr in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
-            self.communicator.broadcast_send(None, sender=sendr)
+        for sender in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
+            self.communicator.broadcast_send(None, sender=sender)
 
         done.result(timeout=self.WAIT_TIMEOUT)
 
@@ -281,7 +320,7 @@ class CommunicatorTester(object):
 
     def test_broadcast_filter_sender_and_subject(self):
         # pylint: disable=invalid-name
-        senders_and_subects = set()
+        senders_and_subjects = set()
         EXPECTED = {
             ('bob.jones', 'purchase.car'),
             ('bob.jones', 'purchase.piano'),
@@ -292,23 +331,75 @@ class CommunicatorTester(object):
         done = kiwipy.Future()
 
         def on_broadcast_1(_communicator, _body, sender=None, subject=None, _correlation_id=None):
-            senders_and_subects.add((sender, subject))
-            if len(senders_and_subects) == len(EXPECTED):
+            senders_and_subjects.add((sender, subject))
+            if len(senders_and_subjects) == len(EXPECTED):
                 done.set_result(True)
 
         filtered = kiwipy.BroadcastFilter(on_broadcast_1)
-        filtered.add_sender_filter("*.jones")
-        filtered.add_subject_filter("purchase.*")
+        filtered.add_sender_filter('*.jones')
+        filtered.add_subject_filter('purchase.*')
         self.communicator.add_broadcast_subscriber(filtered)
 
-        for sendr in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
+        for sender in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
             for subj in ['purchase.car', 'purchase.piano', 'sell.guitar', 'sell.house']:
-                self.communicator.broadcast_send(None, sender=sendr, subject=subj)
+                self.communicator.broadcast_send(None, sender=sender, subject=subj)
 
         done.result(timeout=self.WAIT_TIMEOUT)
 
-        self.assertEqual(4, len(senders_and_subects))
-        self.assertSetEqual(EXPECTED, senders_and_subects)
+        self.assertEqual(4, len(senders_and_subjects))
+        self.assertSetEqual(EXPECTED, senders_and_subjects)
+
+    def test_broadcast_filter_sender_and_subject_regex(self):
+        """As the standard broadcast test but using regular expressions instead of wildcards"""
+        # pylint: disable=invalid-name
+        senders_and_subjects = set()
+        EXPECTED = {
+            ('bob.jones', 'purchase.car'),
+            ('bob.jones', 'purchase.piano'),
+            ('alice.jones', 'purchase.car'),
+            ('alice.jones', 'purchase.piano'),
+        }
+
+        done = kiwipy.Future()
+
+        def on_broadcast_1(_communicator, _body, sender=None, subject=None, _correlation_id=None):
+            senders_and_subjects.add((sender, subject))
+            if len(senders_and_subjects) == len(EXPECTED):
+                done.set_result(True)
+
+        filtered = kiwipy.BroadcastFilter(on_broadcast_1)
+        filtered.add_sender_filter(re.compile('.*.jones'))
+        filtered.add_subject_filter(re.compile('purchase.*'))
+        self.communicator.add_broadcast_subscriber(filtered)
+
+        for sender in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
+            for subj in ['purchase.car', 'purchase.piano', 'sell.guitar', 'sell.house']:
+                self.communicator.broadcast_send(None, sender=sender, subject=subj)
+
+        done.result(timeout=self.WAIT_TIMEOUT)
+
+        self.assertSetEqual(EXPECTED, senders_and_subjects)
+
+    def test_broadcast_filter_match(self):
+        """Test exact match broadcast filter"""
+        EXPECTED_SENDERS = ['alice.jones']  # pylint: disable=invalid-name
+        senders = []
+
+        done = kiwipy.Future()
+
+        def on_broadcast_1(_comm, _body, sender=None, _subject=None, _correlation_id=None):
+            senders.append(sender)
+            if len(senders) == len(EXPECTED_SENDERS):
+                done.set_result(True)
+
+        self.communicator.add_broadcast_subscriber(kiwipy.BroadcastFilter(on_broadcast_1, sender='alice.jones'))
+
+        for sender in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
+            self.communicator.broadcast_send(None, sender=sender)
+
+        done.result(timeout=self.WAIT_TIMEOUT)
+
+        self.assertListEqual(EXPECTED_SENDERS, senders)
 
     def test_add_remove_broadcast_subscriber(self):
         """Test adding, removing and readding a broadcast subscriber"""
@@ -337,6 +428,6 @@ class CommunicatorTester(object):
     def test_broadcast_noone_listening(self):
         """ Check that it's not an error to send a broadcast if no one is listening"""
         result = self.communicator.broadcast_send('into the ether')
-        self.assertIs(True, result)
+        assert result is True
 
     # endregion
